import aiohttp
from .params import Params
from .exceptions import EconetHTTPException, EconetUnauthorized, EconetUnknownId


class Param(object):
    def __init__(self, pid, data, enums):
        self.desc = data.get("name", "")
        self.special = data.get("special", None)
        self.edit = data.get("edit", False)
        self.id = pid
        self.unit = data["unit"]
        self._value = data.get("value", None)
        self.minv = data.get("minv", None)
        self.maxv = data.get("maxv", None)
        self.offset = data.get("offset", None)
        self.mult = data.get("mult", None)
        self.enums = enums

    @property
    def value(self):
        if self.special is not None and self.special != 0 and self._value is not None:
            if len(self.enums[self.special]["values"]) > self._value:
                return self.enums[self.special]["values"][self._value]
        return self._value

    @value.setter
    def value(self, new):
        self._value = new

    def __str__(self):
        return f"{self.value}{self.unit if self.unit is not None else ''}"

    def __repr__(self):
        return f"{self.__class__.__name__}(pid={repr(self.id)})"


class Econet(object):
    def __init__(self, host, login, password):
        self.login = login
        self.password = password
        self.host = host
        self.params = dict()
        self.units = dict()
        self.enums = dict()

    def __repr__(self):
        return f"{self.__class__.__name__}(host={repr(self.host)}, login={repr(self.login)})"

    async def setup(self):
        await self.fetch_units()
        await self.fetch_enums()
        await self.fetch_current_data_params()
        await self.fetch_rm_params()

    async def _call_api(self, cmd):
        async with aiohttp.ClientSession() as session:
            auth = aiohttp.BasicAuth(self.login, self.password)
            async with session.get(f"{self.host}/econet/{cmd}", auth=auth) as resp:
                if resp.status == 401:
                    raise EconetUnauthorized()
                if resp.status != 200:
                    raise EconetHTTPException(
                        f"Got {resp.status} {resp.reason} when calling "
                        f"{self.host}/econet/{cmd}"
                    )
                j = await resp.json()
                return j

    async def fetch_units(self):
        resp = await self._call_api("rmParamsUnitsNames")
        self.units = resp["data"]

    async def fetch_enums(self):
        resp = await self._call_api("rmParamsEnums")
        self.enums = resp["data"]

    async def fetch_current_data_params(self):
        resp = await self._call_api("rmCurrentDataParams")
        for pid, data in resp["data"].items():
            data["unit"] = (
                self.units[int(data["unit"])]
                if int(data["unit"]) < len(self.units)
                else ""
            )
            if pid == "123":
                pid = "110"
            pid = Params.get_by_id(pid)
            self.params[pid] = Param(pid, data, self.enums)

    async def fetch_rm_params(self):
        resp = await self._call_api("rmParamsData")
        for pid, data in enumerate(resp["data"]):
            data["unit"] = (
                self.units[int(data["unit"])]
                if int(data["unit"]) < len(self.units)
                else ""
            )
            pid = Params.get_by_id(pid)
            self.params[pid] = Param(pid, data, self.enums)

    async def fetch_sys_params(self):
        return await self._call_api("sysParams")

    async def fetch_reg_params_data(self):
        resp = await self._call_api("regParamsData")
        for pid, data in resp["data"].items():
            PID = Params.get_by_id(pid)
            if PID is not None:
                self.params[PID].value = data

    async def fetch_rm_params_data(self):
        resp = await self._call_api("rmParamsData")
        for pid, data in enumerate(resp["data"]):
            PID = Params.get_by_id(pid)
            if PID is not None:
                self.params[PID].value = data["value"]

    async def update(self):
        await self.fetch_reg_params_data()
        await self.fetch_rm_params_data()

    def get_param(self, param):
        if (p := self.params.get(param)) is None:
            if type(param) != Params and Params.get_by_id(param) is None:
                raise EconetUnknownId(f"Unknow Parameter ID {param}.")
            undefined_data = dict(name=f"Undefined_{param}", unit="", value="Undefined")
            p = Param(param, undefined_data, dict())
        return p

    async def set_param(self, param_id, value):
        if type(param_id) == int:
            endpoint = f"rmNewParam?newParamIndex={param_id}&newParamValue={value}"
        else:
            endpoint = f"rmCurrNewParam?newParamKey={param_id}&newParamValue={value}"
        resp = await self._call_api(endpoint)
        if resp["result"] != "OK":
            raise Exception(f"Something went wrong: {resp}")



import asyncio
from .params import Params
from .econet import Econet
from .exceptions import EconetUninitilized


class Sensor(object):
    """
    Simple piece of information.
    """

    def __init__(self, econet: Econet, param_id: Params):
        self._param_name = param_id._name_
        self._param = econet.get_param(param_id)
        self._econet = econet

    def __str__(self):
        return f"{str(self.value)}{str(self.unit)}"

    def __repr__(self):
        return (
            f"{repr(self.__class__.__name__)}"
            f"(econet={repr(self._econet)}, param_id={repr(self._param.id)})"
        )

    @property
    def value(self) -> int | str:
        return self._param.value

    @property
    def id(self) -> int | str:
        return self._param.id

    @property
    def unit(self) -> str:
        return self._param.unit

    @property
    def minv(self) -> int:
        return self._param.minv

    @property
    def maxv(self) -> int:
        return self._param.maxv


class Setting(Sensor):
    """
    Represents single setting of the Econet device
    """

    async def set_value(self, value: int) -> int:
        if self.minv is not None and self.maxv is not None:
            if value < self.minv or value > self.maxv:
                raise ValueError(
                    f"Value {value} not in boundries <{self.minv},{self.maxv}>"
                    f" for Setting {self._param_name}"
                )

        task = asyncio.create_task(self._econet.set_param(self._param.id, value))
        await task
        return self.value


class TempController(object):
    """
    Class that controlls one aspect of a furnace, for example boiler temperature or huw temperature.
    """

    def __init__(
        self, econet: Econet, current_temp: Params, target_temp: Params
    ) -> None:
        self._econet = econet
        self._current_temp = Sensor(econet, current_temp)
        self._target_temp = Setting(econet, target_temp)

    @property
    def temperature(self) -> int:
        """Current temeperature."""
        return self._current_temp.value

    @property
    def target_temperature(self) -> int:
        """Current preset value for temperature."""
        return self._target_temp.value

    async def set_target_temperature(self, value: int) -> int:
        """Sets target temperate to desired value using Econet API."""
        resp = await self._target_temp.set_value(value)
        return resp

    @property
    def target_temperature_min(self) -> int:
        """Minimum value when setting target temperature."""
        return self._target_temp.minv

    @property
    def target_temperature_max(self) -> int:
        """Maximum value when setting target temperature."""
        return self._target_temp.maxv


class Smartfire(object):
    def __init__(self, host: str, username: str, password: str):
        self.host = host
        self.username = username
        self.password = password
        self.sensors = dict()

    @property
    def econet(self) -> Econet:
        if hasattr(self, "_econet") is False:
            raise EconetUninitilized
        return self._econet

    @property
    def boiler(self) -> TempController:
        if (sensor := self.sensors.get("boiler")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def huw(self) -> TempController:
        if (sensor := self.sensors.get("huw")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def central_heating(self) -> TempController:
        if (sensor := self.sensors.get("central_heating")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def operation_mode(self) -> str:
        if (sensor := self.sensors.get("operation_mode")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def fuel_level(self) -> str:
        if (sensor := self.sensors.get("fuel_level")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def fan(self) -> str:
        if (sensor := self.sensors.get("fan")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def fan_speed(self) -> str:
        if (sensor := self.sensors.get("fan_speed")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def feeder_temperature(self) -> str:
        if (sensor := self.sensors.get("feeder_temperature")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def mixer1_valve(self) -> str:
        if (sensor := self.sensors.get("mixer1_valve")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def oxygen(self) -> str:
        if (sensor := self.sensors.get("oxygen")) is None:
            raise EconetUninitilized
        return sensor

    @property
    def attributes(self) -> dict:
        return self._attrs

    async def update(self) -> None:
        if hasattr(self, "_econet") is False:
            self._econet = Econet(self.host, self.username, self.password)
            await self._econet.setup()
            self.sensors["boiler"] = TempController(
                self._econet,
                Params.BOILER_TEMPERATURE,
                Params.EDIT_PRESET_BOILER_TEMPERATURE,
            )
            self.sensors["huw"] = TempController(
                self._econet, Params.HUW_TEMPERATURE, Params.EDIT_HUW_PRESET_TEMPERATURE
            )
            self.sensors["central_heating"] = TempController(
                self._econet, Params.TEMP_MIXER_1, Params.PRESET_MIXER_1
            )
            self.sensors["operation_mode"] = Sensor(self._econet, Params.OPERATION_MODE)
            self.sensors["fuel_level"] = Sensor(self._econet, Params.FUEL_LEVEL)
            self.sensors["fan"] = Sensor(self._econet, Params.FAN)
            self.sensors["fan_speed"] = Sensor(
                self._econet, Params.MINIMUM_AIRFLOW_OUTPUT
            )
            self.sensors["oxygen"] = Sensor(self._econet, Params.OXYGEN_2)
            self.sensors["mixer1_valve"] = Sensor(self._econet, Params.VALVE_MIXER_1)
            self.sensors["feeder_temperature"] = Sensor(
                self._econet, Params.FEEDER_TEMPERATURE
            )
            sys = await self._econet.fetch_sys_params()
            self._attrs = {
                "uid": sys["uid"],
                "ecosrvSoftVer": sys["ecosrvSoftVer"],
                "modulePanelSoftVer": sys["modulePanelSoftVer"],
                "moduleASoftVer": sys["moduleASoftVer"],
                "controllerID": sys["controllerID"],
                "settingsVer": sys["settingsVer"],
            }
        await self._econet.update()



from enum import Enum


class Params(Enum):
    def get_by_id(pid):
        return Params._value2member_map_.get(pid, None)

    def __repr__(self):
        return f"{self.__class__.__name__}.{self._name_}"

    LIGHTER = "1"
    POKER = "3"
    FUEL_LEVEL = "110"
    CAUTION_FUEL_LEVEL = "113"
    CAUTION_LOAD_THE_TANK = "114"
    LAMBDA_SENSOR_CALIBRATION_IN_PROGRESS = "116"
    LAMBDA_ERROR = "117"
    DAMAGED_FUEL_SUPPLY_SYSTEM = "118"
    UNSEAL = "122"
    BOILER_THERMOSTAT = "131"
    ROOM_THERMOSTAT_MIXER_1 = "132"
    ROOM_THERMOSTAT_MIXER_2 = "133"
    ROOM_THERMOSTAT_MIXER_3 = "134"
    ROOM_THERMOSTAT_MIXER_4 = "135"
    VALVE_MIXER_1 = "150"
    VALVE_MIXER_2 = "151"
    VALVE_MIXER_3 = "152"
    VALVE_MIXER_4 = "153"
    SERVO_MIXER_1 = "154"
    SERVO_MIXER_2 = "155"
    SERVO_MIXER_3 = "156"
    SERVO_MIXER_4 = "157"
    VALVE_MIXER_1_2 = "158"
    VALVE_MIXER_2_2 = "159"
    VALVE_MIXER_3_2 = "160"
    VALVE_MIXER_4_2 = "161"
    OXYGEN = "162"
    OXYGEN_2 = "165"
    WORK_AT_100 = "166"
    WORK_AT_50 = "167"
    WORK_AT_30 = "168"
    FEEDER_WORK = "169"
    FIRING_UP_COUNT = "170"
    NUMBER_OF_RESETS = "171"
    ETH_IP = "172"
    ETH_MASK = "173"
    ETH_GATEWAY = "174"
    ETH_SERVER_CONNECTED = "175"
    WIFI_IP = "176"
    WIFI_MASK = "177"
    WIFI_GATEWAY = "178"
    WIFI_SERVER_CONNECTED = "179"
    WIFI_ENCRYPTION = "180"
    WIFI_SIGNAL_STRENGHT = "181"
    WIFI_STATUS = "182"
    WIFI_SSID = "184"
    BOILER_TEMPERATURE = "1024"
    HUW_TEMPERATURE = "1025"
    FEEDER_TEMPERATURE = "1026"
    WEATHER_TEMPERATURE = "1027"
    UPPER_BUFFER_TEMPERATURE = "1028"
    LOWER_BUFFER_TEMPERATURE = "1029"
    EMISSION_TEMPERATURE = "1030"
    TEMP_MIXER_1 = "1031"
    TEMP_MIXER_2 = "1032"
    TEMP_MIXER_3 = "1033"
    TEMP_MIXER_4 = "1034"
    PRESET_BOILER_TEMPERATURE = "1280"
    HUW_PRESET_TEMPERATURE = "1281"
    PRESET_TEMP_MIXER_1 = "1287"
    PRESET_TEMP_MIXER_2 = "1288"
    PRESET_TEMP_MIXER_3 = "1289"
    PRESET_TEMP_MIXER_4 = "1290"
    FAN = "1536"
    FEEDER = "1538"
    ADDITIONAL_FEEDER = "1540"
    BOILER_PUMP = "1541"
    HUW_PUMP = "1542"
    CIRCULATING_PUMP = "1543"
    PUMP_MIXER_1 = "1544"
    PUMP_MIXER_2 = "1547"
    PUMP_MIXER_3 = "1550"
    PUMP_MIXER_4 = "1553"
    OPERATION_MODE = "1792"
    BURNER_OUTPUT = "1794"
    EDIT_PRESET_BOILER_TEMPERATURE = 0
    MIN_BOILER_TEMPERATURE = 1
    MAX_BOILER_TEMPERATURE = 2
    BOILER_INC_BY_HUW_MIXER = 3
    CH_PUMP_ACTIVATION_TEMPERATURE = 4
    DHW_PUMP = 5
    HUW_PUMP_MODE = 8
    BLOW_IN_OUTPUT_100 = 10
    FEEDER_OPERATION_100 = 11
    FEEDER_INTERVAL_100 = 12
    FEEDER_WORK_CORRECTION_50 = 13
    FEEDER_OPERATION_50 = 14
    FEEDER_INTERVAL_50 = 15
    BLOW_IN_OUTPUT_30 = 16
    FEEDER_OPERATION_30 = 17
    FEEDER_INTERVAL_30 = 18
    H2_HYSTERESIS_50 = 19
    H1_HYSTERESIS_30 = 20
    BOILER_HYSTERESIS = 21
    FIRING_UP_AIRFLOW = 22
    IGNITION_TEST_TIME = 23
    IGNITION_TEST_TIME_2 = 24
    FEEDING_TIME = 25
    FIRING_UP_TIME = 26
    EX_TEMP_DELTA = 27
    EX_TEMP_DELTA_2 = 28
    EX_TEMP_AT_THE_END_OF_FIRING_UP = 29
    TEST_DOSE = 30
    SUPERVISION_TIME = 31
    FEED_INTERVAL = 32
    AIRFLOW_OPER_EXTEND = 33
    MINIMUM_AIRFLOW_OUTPUT = 34
    STABILIZATION_TIME = 35
    BURNING_OFF_TIME = 36
    AIR_FLUSH_INTENSITY = 37
    OXYGEN_100 = 38
    OXYGEN_50 = 39
    OXYGEN_30 = 40
    AIRFLOW_CORRECTION_RANGE = 41
    POKER_CYCLE_TIME = 42
    EX_TEMP_W_NO_FUEL = 43
    EDIT_HUW_PRESET_TEMPERATURE = 47
    MINIMUM_HUW_TEMPERATURE = 48
    MAXIMUM_HUW_TEMPERATURE = 49
    HUW_CONT_HYSTERESIS = 50
    HUW_DISINFECTION = 52
    SUMMER_MODE = 56
    CH_PUMP_STANDSTILL_WHEN_LOADING_HUW = 59
    EXTENDING_HUW_PUMP_OPERATION_TIME = 60
    EXCHANGER = 61
    PRESET_MIXER_1 = 62
    MIN_MIXER_1_TEMP = 66
    MAX_MIXER_1_TEMP = 70
    MIXER_1_ROOM_THERM = 74
    MIXER_1_SUPPORT = 86
    PROPORTIONAL_RANGE = 90
    INTEGRATION_TIME_CONSTANT = 94
    OFF_BY_THERMOSTAT = 98
    VALVE_FULL_OPENING_TIME = 102
    MIXER_1_INPUT_DEAD_ZONE = 106
    BOILER_COOLING_TEMPERATURE = 111
    AIRFLOW_POWER_CORRECTION_100 = 129
    FEEDER_WORK_CORRECTION_100 = 130
    AIRFLOW_POWER_CORRECTION_50 = 131
    FEEDER_WORK_CORRECTIN_50 = 132
    AIRFLOW_POWER_CORRECTION_30 = 133
    FEEDER_WORK_CORRECTION_30 = 134
    BURNER_CLEANING = 135
    CLEANING_SYSTEM_OPERATION_TIME = 136
    ALARM_LEVEL = 137
    NO_FUEL_DETECTION_TIME = 138
    FEEDER_OPERAT_TIME_2 = 140
    FEEDER_INTERVAL_2 = 141
    FEEDER_2_MESSAGE = 142
    MAX_BURNER_TEMP = 143
    OPERATION_WITH_LAMBDA_SENSOR = 144
    FEEDER_LOCK = 145
    FUEL_DETECTION_OXYGEN = 146
    FUEL_DETECTION_TIME = 147
    PARAMETER_A = 148
    PARAMETER_B = 149
    PARAMETER_C = 150
    OUTPUT_H1 = 152
    H2_OUTPUT = 153
    HEAT_SOURCE = 157
    THERMOSTAT_LOCK = 159
    BOILER_PUMP_LOCK = 160
    SHOW_ADVANCED_SETUP = 164
    ALARMS = 165
